# coding: utf-8
from typing import List, Optional, Dict, Any
from file_state_manager.cloneable_file import CloneableFile
from delta_trace_db.query.cause.temporal_trace.timestamp_node import TimestampNode

class TemporalTrace(CloneableFile):
    """イベントの「時間の軌跡」を記録するクラス。"""
    class_name = "TemporalTrace"
    version = "1"

    def __init__(self, nodes: Optional[List[TimestampNode]] = None):
        super().__init__()
        self.nodes: List[TimestampNode] = nodes if nodes is not None else []

    @property
    def initiated_at(self) -> Optional[str]:
        """最初のイベント発生時刻。"""
        return self.nodes[0].timestamp if self.nodes else None

    @property
    def finalized_at(self) -> Optional[str]:
        """最後のイベント記録時刻。"""
        return self.nodes[-1].timestamp if self.nodes else None

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "TemporalTrace":
        nodes_list = src.get("nodes", [])
        nodes = [TimestampNode.from_dict(n) for n in nodes_list]
        return cls(nodes=nodes)

    def clone(self) -> "TemporalTrace":
        return TemporalTrace.from_dict(self.to_dict())

    def to_dict(self) -> Dict[str, Any]:
        return {
            "className": self.class_name,
            "version": self.version,
            "nodes": [node.to_dict() for node in self.nodes],
        }



# coding: utf-8
import json
from datetime import datetime
from copy import deepcopy

from file_state_manager.cloneable_file import CloneableFile
from delta_trace_db.db.delta_trace_db_core import DeltaTraceDatabase
from delta_trace_db.query.nodes.comparison_node import FieldStartsWith, FieldEquals, FieldGreaterThan
from delta_trace_db.query.nodes.logical_node import OrNode
from delta_trace_db.query.query_builder import QueryBuilder
from delta_trace_db.query.query_result import QueryResult
from delta_trace_db.query.sort.single_sort import SingleSort


class User(CloneableFile):
    def __init__(self, id_: str, name: str, age: int, created_at: datetime, updated_at: datetime, nested_obj: dict):
        super().__init__()
        self.id = id_
        self.name = name
        self.age = age
        self.created_at = created_at
        self.updated_at = updated_at
        self.nested_obj = nested_obj

    @classmethod
    def from_dict(cls, src: dict):
        return cls(
            id_=src['id'],
            name=src['name'],
            age=src['age'],
            created_at=datetime.fromisoformat(src['createdAt']),
            updated_at=datetime.fromisoformat(src['updatedAt']),
            nested_obj=deepcopy(src['nestedObj'])
        )

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'age': self.age,
            'createdAt': self.created_at.isoformat(),
            'updatedAt': datetime.now().isoformat(),
            'nestedObj': deepcopy(self.nested_obj)
        }

    def clone(self):
        return User.from_dict(self.to_dict())


class User2(CloneableFile):
    def __init__(self, id_: str, name: str, age: int, nested_obj: dict, gender: str):
        super().__init__()
        self.id = id_
        self.name = name
        self.age = age
        self.nested_obj = nested_obj
        self.gender = gender

    @classmethod
    def from_dict(cls, src: dict):
        return cls(
            id_=src['id'],
            name=src['name'],
            age=src['age'],
            nested_obj=deepcopy(src['nestedObj']),
            gender=src['gender']
        )

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'age': self.age,
            'nestedObj': deepcopy(self.nested_obj),
            'gender': self.gender
        }

    def clone(self):
        return User2.from_dict(self.to_dict())


class Item1(CloneableFile):
    def __init__(self, name, serial_key=None):
        super().__init__()
        self.serial_key = serial_key
        self.name = name

    @classmethod
    def from_dict(cls, src) -> "Item1":
        return Item1(
            serial_key=src["serialKey"],
            name=src["name"],
        )

    def to_dict(self):
        return {
            "serialKey": self.serial_key,
            "name": self.name,
        }

    def clone(self):
        return Item1.from_dict(self.to_dict())

# --------------------------
# Test function
# --------------------------

def test_speed_crud():
    records_count = 100000
    print(f"speed test for {records_count} records")
    now = datetime.now()

    db = DeltaTraceDatabase()
    users = [
        User(
            id_=str(i),
            name=f'sample{i}',
            age=i,
            created_at=now,
            updated_at=now,
            nested_obj={"num": i}
        )
        for i in range(records_count)
    ]

    # add
    q1 = QueryBuilder.add(target='users', add_data=users).build()
    print("start add")
    dt1 = datetime.now()
    r1: QueryResult = db.execute_query(q1)
    dt2 = datetime.now()
    print(f"end add: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")
    assert r1.is_success

    # getAll
    q1_get = QueryBuilder.get_all(target='users').build()
    print("start getAll (with object convert)")
    dt1 = datetime.now()
    r1_get: QueryResult = db.execute_query(q1_get)
    _ = r1_get.convert(User.from_dict)
    dt2 = datetime.now()
    print(f"end getAll: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")
    assert r1.is_success
    print(f"returnsLength: {len(r1_get.result)}")

    # save
    print("start save (with json string convert)")
    dt1 = datetime.now()
    db_map = db.to_dict()
    json_str = json.dumps(db_map)
    dt2 = datetime.now()
    print(f"end save: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")

    # load
    print("start load (with json string convert)")
    dt1 = datetime.now()
    json_map = json.loads(json_str)
    _ = DeltaTraceDatabase.from_dict(json_map)
    dt2 = datetime.now()
    print(f"end load: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")

    # search
    q2 = QueryBuilder.search(
        target='users',
        query_node=FieldStartsWith("name", "sample"),
        sort_obj=SingleSort(field='age')
    ).build()
    print("start search (with object convert)")
    dt1 = datetime.now()
    r2: QueryResult = db.execute_query(q2)
    _ = r2.convert(User.from_dict)
    dt2 = datetime.now()
    print(f"end search: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")
    print(f"returnsLength: {len(r2.result)}")

    # search paging
    q2_paging = QueryBuilder.search(
        target='users',
        query_node=FieldStartsWith("name", "sample"),
        sort_obj=SingleSort(field='age'),
        limit=records_count // 2
    ).build()
    print("start search paging, half limit pre search (with object convert)")
    dt1 = datetime.now()
    r2_paging: QueryResult = db.execute_query(q2_paging)
    _ = r2_paging.convert(User.from_dict)
    dt2 = datetime.now()
    print(f"end search paging: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")
    print(f"returnsLength: {len(r2_paging.result)}")

    # search one
    q2_one = QueryBuilder.search_one(
        target='users',
        query_node=FieldEquals('age', records_count - 1),
    ).build()
    print("start searchOne, the last index object search (with object convert)")
    dt1 = datetime.now()
    r2_one: QueryResult = db.execute_query(q2_one)
    _ = r2_one.convert(User.from_dict)
    dt2 = datetime.now()
    print(f"end searchOne: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")
    print(f"returnsLength: {len(r2_one.result)}")

    # update
    q3 = QueryBuilder.update(
        target='users',
        query_node=OrNode([
            FieldEquals('name', f'sample{(records_count // 2) -1}'),
            FieldEquals('name', f'sample{records_count - 1}')
        ]),
        override_data={'age': records_count + 1},
        return_data=False,
        sort_obj=SingleSort(field='id')
    ).build()
    print("start update at half index and last index object")
    dt1 = datetime.now()
    _ = db.execute_query(q3)
    dt2 = datetime.now()
    print(f"end update: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")

    # updateOne
    q4 = QueryBuilder.update_one(
        target='users',
        query_node=FieldEquals('name', f'sample{(records_count // 2 )-1}'),
        override_data={'age': records_count},
        return_data=False
    ).build()
    print("start updateOne of half index object")
    dt1 = datetime.now()
    _ = db.execute_query(q4)
    dt2 = datetime.now()
    print(f"end updateOne: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")

    # conformToTemplate
    q5 = QueryBuilder.conform_to_template(
        target='users',
        template=User2(
            id_="None",
            name="None",
            age=-1,
            nested_obj={"num": -1},
            gender="None"
        ).to_dict()
    ).build()
    print("start conformToTemplate")
    db2 = db.clone()
    dt1 = datetime.now()
    _ = db2.execute_query(q5)
    dt2 = datetime.now()
    print(f"end conformToTemplate: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")

    # delete
    q6 = QueryBuilder.delete(
        target='users',
        query_node=FieldGreaterThan('age', (records_count // 2)),
        sort_obj=SingleSort(field='id'),
        return_data=True
    ).build()
    print("start delete half object (with object convert)")
    dt1 = datetime.now()
    r6 = db.execute_query(q6)
    _ = r6.convert(User.from_dict)
    dt2 = datetime.now()
    print(f"end delete: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")
    print(f"returnsLength: {len(r6.result)}")

    # deleteOne
    q7 = QueryBuilder.delete_one(
        target='users',
        query_node=FieldEquals('age', (records_count // 2)),
        return_data=True
    ).build()
    print("start deleteOne for last object (with object convert)")
    dt1 = datetime.now()
    r7 = db.execute_query(q7)
    _ = r7.convert(User.from_dict)
    dt2 = datetime.now()
    print(f"end deleteOne: {(dt2 - dt1).total_seconds() * 1000:.0f} ms")
    print(f"returnsLength: {len(r7.result)}")

    # add with serialKey
    items = []
    for i in range(records_count):
        items.append(Item1(name=str(i)))
    q8 = QueryBuilder.add(target="items", add_data=items, serial_key="serialKey").build()
    print("start add with serialKey")
    dt1 = datetime.now()
    r8: QueryResult = db.execute_query(q8)
    dt2 = datetime.now()
    elapsed_ms = int((dt2.timestamp() - dt1.timestamp()) * 1000)
    print(f"end add with serialKey: {elapsed_ms} ms")
    assert r8.is_success is True
    print("addedCount:" + str(r8.db_length))



# coding: utf-8
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict


class AbstractSort(ABC):
    """Base class for sort objects.

    (en) Comparison function for sorting.
    (ja) ソート用の比較関数です。
    """

    @abstractmethod
    def get_comparator(self) -> Callable[[Dict[str, Any], Dict[str, Any]], int]:
        """Return a comparator function for sorting."""
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """Convert the object to a dictionary."""
        pass

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "AbstractSort":
        """Restore a sort object from a dictionary."""
        # 遅延インポート
        from delta_trace_db.query.sort.single_sort import SingleSort
        from delta_trace_db.query.sort.multi_sort import MultiSort
        class_name = src.get("className")
        if class_name == SingleSort.class_name:
            return SingleSort.from_dict(src)
        elif class_name == MultiSort.class_name:
            return MultiSort.from_dict(src)
        else:
            raise ValueError("Unknown sort class")



# coding: utf-8
# logical_node.py
from typing import Any, Dict
from delta_trace_db.query.nodes.enum_node_type import EnumNodeType
from delta_trace_db.query.nodes.query_node import QueryNode


class AndNode(QueryNode):
    def __init__(self, conditions: list[QueryNode]):
        self.conditions = conditions

    @classmethod
    def from_dict(cls, src: Dict[str, Any]):
        from delta_trace_db.query.nodes.query_node import QueryNode
        return cls([QueryNode.from_dict(dict(e)) for e in src['conditions']])

    def evaluate(self, data: dict) -> bool:
        return all(c.evaluate(data) for c in self.conditions)

    def to_dict(self) -> dict:
        return {
            'type': EnumNodeType.and_.name,
            'conditions': [c.to_dict() for c in self.conditions],
            'version': '1',
        }


class OrNode(QueryNode):
    def __init__(self, conditions: list[QueryNode]):
        self.conditions = conditions

    @classmethod
    def from_dict(cls, src: dict):
        from delta_trace_db.query.nodes.query_node import QueryNode
        return cls([QueryNode.from_dict(dict(e)) for e in src['conditions']])

    def evaluate(self, data: dict) -> bool:
        return any(c.evaluate(data) for c in self.conditions)

    def to_dict(self) -> dict:
        return {
            'type': EnumNodeType.or_.name,
            'conditions': [c.to_dict() for c in self.conditions],
            'version': '1',
        }


class NotNode(QueryNode):
    def __init__(self, condition):
        self.condition = condition

    @classmethod
    def from_dict(cls, src: dict):
        from delta_trace_db.query.nodes.query_node import QueryNode
        return cls(QueryNode.from_dict(dict(src['condition'])))

    def evaluate(self, data: dict) -> bool:
        return not self.condition.evaluate(data)

    def to_dict(self) -> dict:
        return {
            'type': EnumNodeType.not_.name,
            'condition': self.condition.to_dict(),
            'version': '1',
        }



from datetime import datetime, timezone, timedelta
from delta_trace_db.db.delta_trace_db_core import DeltaTraceDatabase
from delta_trace_db.query.nodes.comparison_node import FieldEquals
from delta_trace_db.query.nodes.enum_value_type import EnumValueType
from delta_trace_db.query.raw_query_builder import RawQueryBuilder
from delta_trace_db.query.sort.single_sort import SingleSort


def test_timezone_search():
    db = DeltaTraceDatabase()

    # --- データ追加 ---
    q1 = RawQueryBuilder.add(
        target="user",
        raw_add_data=[
            {"id": "0", "time": "2025-10-11T18:30:00+09:00"},  # JST
            {"id": "1", "time": "2025-10-11T18:30:00+09:00"},  # JST
            {"id": "2", "time": "2025-10-11T09:30:00Z"},  # UTC
        ],
    ).build()
    db.execute_query(q1)

    local_time = datetime(2025, 10, 11, 18, 30)
    # --- UTC（タイムゾーン付き）でソートありで検索 ---
    jst = timezone(timedelta(hours=9))
    local_aware = local_time.replace(tzinfo=jst)  # 2025-10-11T18:30+09:00
    utc_time = local_aware.astimezone(timezone.utc)  # -> 2025-10-11T09:30+00:00
    q2 = RawQueryBuilder.search(target="user", query_node=FieldEquals(field="time", value=utc_time),
                                sort_obj=SingleSort(field="time", v_type=EnumValueType.datetime_)).build()
    r2 = db.execute_query(q2)

    assert r2.is_success is True
    assert len(r2.result) == 3
    ids = [r["id"] for r in r2.result]
    assert ids == ["0", "1", "2"]



# coding: utf-8
from delta_trace_db.db.delta_trace_db_core import DeltaTraceDatabase
from delta_trace_db.query.raw_query_builder import RawQueryBuilder
from delta_trace_db.query.transaction_query import TransactionQuery


def test_remove_collection():
    db = DeltaTraceDatabase()

    # q1: add
    q1 = RawQueryBuilder.add(
        target="user",
        raw_add_data=[
            {"id": -1, "name": "a"},
            {"id": -1, "name": "a"},
        ],
        serial_key="id",
    ).build()
    db.execute_query(q1)
    assert db.collection("user").length == 2

    # q2: removeCollection
    q2 = RawQueryBuilder.remove_collection(target="user").build()
    r2 = db.execute_query(q2)
    assert r2.is_success is True
    assert db.find_collection("user") is None

    # q3: removeCollection with must_affect_at_least_one=True
    q3 = RawQueryBuilder.remove_collection(
        target="user",
        must_affect_at_least_one=True,
    ).build()
    r3 = db.execute_query(q3)
    assert r3.is_success is False
    assert db.find_collection("user") is None

    # Transaction not permitted
    tq1 = TransactionQuery(queries=[q3])
    tr1 = db.execute_transaction_query(tq1)
    assert tr1.is_success is False
    assert (
        tr1.error_message
        == "The query contains a type that is not permitted to be executed within a transaction."
    )



# coding: utf-8
from delta_trace_db.db.delta_trace_db_core import DeltaTraceDatabase
from delta_trace_db.query.nodes.comparison_node import FieldEquals
from delta_trace_db.query.raw_query_builder import RawQueryBuilder


def test_search_one():
    db = DeltaTraceDatabase()

    # q1: add
    q1 = RawQueryBuilder.add(
        target="user",
        raw_add_data=[
            {"id": -1, "name": "a"},
            {"id": -1, "name": "a"},
        ],
        serial_key="id",
    ).build()
    db.execute_query(q1)
    assert db.collection("user").length == 2

    # q2: searchOne
    q2 = RawQueryBuilder.search_one(
        target="user",
        query_node=FieldEquals("name", "a"),
    ).build()
    r2 = db.execute_query(q2)
    assert r2.is_success is True
    assert len(r2.result) == 1
    assert r2.hit_count == 1
    assert r2.result[0]["id"] == 0



# coding: utf-8
from typing import Any, Dict, List, Optional
from delta_trace_db.query.enum_query_type import EnumQueryType
from delta_trace_db.query.nodes.query_node import QueryNode
from delta_trace_db.query.sort.abstract_sort import AbstractSort
from delta_trace_db.query.cause.cause import Cause
from delta_trace_db.db.util_copy import UtilCopy


class Query:
    className: str = "Query"
    version: str = "6"

    def __init__(
            self,
            target: str,
            type_: EnumQueryType,
            add_data: Optional[List[Dict[str, Any]]] = None,
            override_data: Optional[Dict[str, Any]] = None,
            template: Optional[Dict[str, Any]] = None,
            query_node: Optional[QueryNode] = None,
            sort_obj: Optional[AbstractSort] = None,
            offset: Optional[int] = None,
            start_after: Optional[Dict[str, Any]] = None,
            end_before: Optional[Dict[str, Any]] = None,
            rename_before: Optional[str] = None,
            rename_after: Optional[str] = None,
            limit: Optional[int] = None,
            return_data: bool = False,
            must_affect_at_least_one: bool = True,
            serial_key: Optional[str] = None,
            reset_serial: bool = False,
            cause: Optional[Cause] = None,
    ):
        self.target = target
        self.type = type_
        self.add_data = add_data
        self.override_data = override_data
        self.template = template
        self.query_node = query_node
        self.sort_obj = sort_obj
        self.offset = offset
        self.start_after = start_after
        self.end_before = end_before
        self.rename_before = rename_before
        self.rename_after = rename_after
        self.limit = limit
        self.return_data = return_data
        self.must_affect_at_least_one = must_affect_at_least_one
        self.serial_key = serial_key
        self.reset_serial = reset_serial
        self.cause = cause

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "Query":
        return cls(
            target=src["target"],
            type_=EnumQueryType[src["type"]],
            add_data=src.get("addData"),
            override_data=src.get("overrideData"),
            template=src.get("template"),
            query_node=QueryNode.from_dict(src["queryNode"])
            if src.get("queryNode")
            else None,
            sort_obj=AbstractSort.from_dict(src["sortObj"])
            if src.get("sortObj")
            else None,
            offset=src.get("offset"),
            start_after=src.get("startAfter"),
            end_before=src.get("endBefore"),
            rename_before=src.get("renameBefore"),
            rename_after=src.get("renameAfter"),
            limit=src.get("limit"),
            return_data=src.get("returnData", False),
            must_affect_at_least_one=src.get("mustAffectAtLeastOne", True),
            serial_key=src.get("serialKey", None),
            reset_serial=src.get("resetSerial", False),
            cause=Cause.from_dict(src["cause"]) if src.get("cause") else None,
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "className": self.className,
            "version": self.version,
            "target": self.target,
            "type": self.type.name,
            "addData": (
                [dict(UtilCopy.jsonable_deep_copy(d)) for d in self.add_data]
                if self.add_data is not None
                else None
            ),
            "overrideData": UtilCopy.jsonable_deep_copy(self.override_data),
            "template": UtilCopy.jsonable_deep_copy(self.template),
            "queryNode": self.query_node.to_dict() if self.query_node else None,
            "sortObj": self.sort_obj.to_dict() if self.sort_obj else None,
            "offset": self.offset,
            "startAfter": UtilCopy.jsonable_deep_copy(self.start_after),
            "endBefore": UtilCopy.jsonable_deep_copy(self.end_before),
            "renameBefore": self.rename_before,
            "renameAfter": self.rename_after,
            "limit": self.limit,
            "returnData": self.return_data,
            "mustAffectAtLeastOne": self.must_affect_at_least_one,
            "serialKey": self.serial_key,
            "resetSerial": self.reset_serial,
            "cause": self.cause.to_dict() if self.cause else None,
        }

    def clone(self) -> "Query":
        return Query.from_dict(self.to_dict())



# coding: utf-8
from datetime import datetime, timezone
from typing import Dict, Any
from file_state_manager.cloneable_file import CloneableFile
from delta_trace_db.db.util_copy import UtilCopy

def to_utc_iso(dt: datetime) -> str:
    """
    Dart の timestamp.toUtc().toIso8601String() と同じ挙動で
    UTC ISO8601 文字列に変換する関数。

    - aware datetime（タイムゾーン付き）の場合は UTC に変換
    - naive datetime（タイムゾーンなし）の場合は UTC として扱う
    """
    if dt.tzinfo is None:
        # naive datetime → UTC として扱う
        dt_utc = dt.replace(tzinfo=timezone.utc)
    else:
        # aware datetime → UTC に変換
        dt_utc = dt.astimezone(timezone.utc)
    return dt_utc.isoformat()

class TimestampNode(CloneableFile):
    """軌跡上の各チェックポイントを表すノード。"""

    class_name = "TimestampNode"
    version = "1.post1"

    def __init__(self, timestamp: datetime, location: str, context: Dict[str, Any] = None):
        super().__init__()
        self.timestamp: datetime = timestamp
        self.location: str = location
        self.context: Dict[str, Any] = context if context is not None else {}

    @classmethod
    def from_dict(cls, src: Dict[str, Any]):
        return cls(
            timestamp=datetime.fromisoformat(src["timestamp"]),
            location=src["location"],
            context=src.get("context", {}),
        )

    def clone(self):
        return TimestampNode.from_dict(self.to_dict())

    def to_dict(self) -> Dict[str, Any]:
        return {
            "className": self.class_name,
            "version": self.version,
            "timestamp": to_utc_iso(dt=self.timestamp),
            "location": self.location,
            "context": UtilCopy.jsonable_deep_copy(self.context),
        }



# coding: utf-8
from threading import RLock
from typing import Any, Dict, List, Callable, Optional

from file_state_manager.cloneable_file import CloneableFile

from delta_trace_db.db.delta_trace_db_collection import Collection
from delta_trace_db.query.cause.permission import Permission
from delta_trace_db.query.enum_query_type import EnumQueryType
from delta_trace_db.query.query import Query
from delta_trace_db.query.query_execution_result import QueryExecutionResult
from delta_trace_db.query.query_result import QueryResult
from delta_trace_db.query.transaction_query import TransactionQuery
from delta_trace_db.query.transaction_query_result import TransactionQueryResult
from delta_trace_db.query.util_query import UtilQuery
import logging

_logger = logging.getLogger(__name__)


class DeltaTraceDatabase(CloneableFile):
    class_name = "DeltaTraceDatabase"
    version = "12"

    def __init__(self):
        super().__init__()
        self._collections: Dict[str, Collection] = {}
        self._lock = RLock()  # execute_query / execute_transaction_query 共通

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "DeltaTraceDatabase":
        instance = cls()
        instance._collections = cls._parse_collections(src)
        return instance

    @staticmethod
    def _parse_collections(src: Dict[str, Any]) -> Dict[str, Collection]:
        cols = src.get("collections")
        if not isinstance(cols, dict):
            raise ValueError("Invalid format: 'collections' should be a dict")
        result: Dict[str, Collection] = {}
        for key, value in cols.items():
            if not isinstance(value, dict):
                raise ValueError("Invalid format: target is not a dict")
            result[key] = Collection.from_dict(value)
        return result

    def collection(self, name: str) -> Collection:
        with self._lock:
            if name in self._collections:
                return self._collections[name]
            col = Collection()
            self._collections[name] = col
            return col

    def find_collection(self, name: str) -> Optional[Collection]:
        """
        (en) Find the specified collection.
        Returns it if it exists, otherwise returns None.

        (ja) 指定のコレクションを検索します。
        存在すれば返し、存在しなければ None を返します。

        * name : The collection name.
        """
        with self._lock:
            return self._collections.get(name)

    def remove_collection(self, name: str) -> None:
        """
        (en) Deletes the specified collection.
        If a collection with the specified name does not exist, this does nothing.

        (ja) 指定のコレクションを削除します。
        指定の名前のコレクションが存在しなかった場合は何もしません。

        * name : The collection name.
        """
        with self._lock:
            self._collections.pop(name, None)

    def collection_to_dict(self, name: str) -> Optional[Dict[str, Any]]:
        with self._lock:
            collection = self._collections.get(name)
            return collection.to_dict() if collection is not None else None

    def collection_from_dict(self, name: str, src: Dict[str, Any]) -> Collection:
        with self._lock:
            col = Collection.from_dict(src)
            self._collections[name] = col
            return col

    def collection_from_dict_keep_listener(self, name: str, src: Dict[str, Any]) -> Collection:
        """
        (en) Restores a specific collection from a dictionary, re-registers it,
        and retrieves it.
        If a collection with the same name already exists, it will be overwritten.
        This is typically used to restore data saved with collection_to_dict.
        This method preserves existing listeners when overwriting the specified
        collection.

        (ja) 特定のコレクションを辞書から復元して再登録し、取得します。
        既存の同名のコレクションが既にある場合は上書きされます。
        通常は、collection_to_dictで保存したデータを復元する際に使用します。
        このメソッドでは、指定されたコレクションの上書き時、既存のリスナが維持されます。

        * name : The collection name.
        * src : A dictionary made with collection_to_dict of this class.
        """
        with self._lock:
            col = Collection.from_dict(src)
            listeners_buf = None
            named_listeners_buf = None
            if name in self._collections:
                listeners_buf = getattr(self._collections[name], "listeners", None)
                named_listeners_buf = getattr(self._collections[name], "named_listeners", None)
            self._collections[name] = col
            if listeners_buf is not None:
                col.listeners = listeners_buf
            if named_listeners_buf is not None:
                col.named_listeners = named_listeners_buf
            return col

    def clone(self) -> "DeltaTraceDatabase":
        with self._lock:
            return DeltaTraceDatabase.from_dict(self.to_dict())

    @property
    def raw(self) -> Dict[str, Collection]:
        return self._collections

    def to_dict(self) -> Dict[str, Any]:
        with self._lock:
            return {
                "className": self.class_name,
                "version": self.version,
                "collections": {k: v.to_dict() for k, v in self._collections.items()},
            }

    def add_listener(self, target: str, cb: Callable[[], None], name: Optional[str] = None):
        with self._lock:
            self.collection(target).add_listener(cb, name=name)

    def remove_listener(self, target: str, cb: Callable[[], None], name: Optional[str] = None):
        with self._lock:
            self.collection(target).remove_listener(cb, name=name)

    def execute_query_object(self, query: Any,
                             collection_permissions: Optional[Dict[str, Permission]] = None) -> QueryExecutionResult:
        with self._lock:  # 排他制御
            if isinstance(query, Query):
                return self.execute_query(query, collection_permissions=collection_permissions)
            elif isinstance(query, TransactionQuery):
                return self.execute_transaction_query(query, collection_permissions=collection_permissions)
            elif isinstance(query, dict):
                if query.get("className") == "Query":
                    return self.execute_query(Query.from_dict(query), collection_permissions=collection_permissions)
                elif query.get("className") == "TransactionQuery":
                    return self.execute_transaction_query(TransactionQuery.from_dict(query),
                                                          collection_permissions=collection_permissions)
                else:
                    raise ValueError("Unsupported query class")
            else:
                raise ValueError("Unsupported query type")

    def execute_query(self, q: Query, collection_permissions: Optional[Dict[str, Permission]] = None) -> QueryResult:
        with self._lock:  # 単体クエリもここで排他
            # パーミッションのチェック
            if not UtilQuery.check_permissions(q=q, collection_permissions=collection_permissions):
                return QueryResult(
                    is_success=False,
                    target=q.target,
                    type_=q.type,
                    result=[],
                    db_length=-1,
                    update_count=0,
                    hit_count=0,
                    error_message="Operation not permitted."
                )
            is_exist_col = q.target in self._collections
            col = self.collection(q.target)
            try:
                match q.type:
                    case EnumQueryType.add:
                        r = col.add_all(q)
                    case EnumQueryType.update:
                        r = col.update(q, is_single_target=False)
                    case EnumQueryType.updateOne:
                        r = col.update(q, is_single_target=True)
                    case EnumQueryType.delete:
                        r = col.delete(q)
                    case EnumQueryType.deleteOne:
                        r = col.delete_one(q)
                    case EnumQueryType.search:
                        r = col.search(q)
                    case EnumQueryType.searchOne:
                        r = col.search_one(q)
                    case EnumQueryType.getAll:
                        r = col.get_all(q)
                    case EnumQueryType.conformToTemplate:
                        r = col.conform_to_template(q)
                    case EnumQueryType.renameField:
                        r = col.rename_field(q)
                    case EnumQueryType.count:
                        r = col.count(q)
                    case EnumQueryType.clear:
                        r = col.clear(q)
                    case EnumQueryType.clearAdd:
                        r = col.clear_add(q)
                    case EnumQueryType.removeCollection:
                        if is_exist_col:
                            r = QueryResult(
                                is_success=True,
                                target=q.target,
                                type_=q.type,
                                result=[],
                                db_length=0,
                                update_count=1,
                                hit_count=0,
                                error_message=None)
                        else:
                            r = QueryResult(
                                is_success=True,
                                target=q.target,
                                type_=q.type,
                                result=[],
                                db_length=0,
                                update_count=0,
                                hit_count=0,
                                error_message=None)
                        self.remove_collection(name=q.target)
                # must_affect_at_least_oneの判定。
                if q.type in (
                        EnumQueryType.add,
                        EnumQueryType.update,
                        EnumQueryType.updateOne,
                        EnumQueryType.delete,
                        EnumQueryType.deleteOne,
                        EnumQueryType.conformToTemplate,
                        EnumQueryType.renameField,
                        EnumQueryType.clear,
                        EnumQueryType.clearAdd,
                        EnumQueryType.removeCollection
                ):
                    if q.must_affect_at_least_one and r.update_count == 0:
                        return QueryResult(
                            is_success=False,
                            target=q.target,
                            type_=q.type,
                            result=[],
                            db_length=len(col.raw),
                            update_count=0,
                            hit_count=r.hit_count,
                            error_message="No data matched the condition (mustAffectAtLeastOne=True)"
                        )
                return r
            except ValueError:
                # ここでは安全なメッセージのみを外部に返す
                _logger.error("execute_query ArgumentError", exc_info=True)
                return QueryResult(
                    is_success=False,
                    target=q.target,
                    type_=q.type,
                    result=[],
                    db_length=len(col.raw),
                    update_count=-1,
                    hit_count=-1,
                    error_message="execute_query ArgumentError"
                )
            except Exception:
                _logger.error("execute_query Unexpected Error", exc_info=True)
                return QueryResult(
                    is_success=False,
                    target=q.target,
                    type_=q.type,
                    result=[],
                    db_length=len(col.raw),
                    update_count=-1,
                    hit_count=-1,
                    error_message="execute_query Unexpected Error",
                )

    def execute_transaction_query(self, q: TransactionQuery,
                                  collection_permissions: Optional[
                                      Dict[str, Permission]] = None) -> TransactionQueryResult:
        with self._lock:  # トランザクション全体で排他
            # 許可されていないクエリが混ざっていないか調査し、混ざっていたら失敗にする。
            for i in q.queries:
                if i.type == EnumQueryType.removeCollection:
                    return TransactionQueryResult(is_success=False, results=[],
                                                  error_message="The query contains a type that is not permitted to be executed within a transaction.")
            # トランザクション付き処理を開始。
            results: List[QueryResult] = []
            try:
                buff: Dict[str, Dict[str, Any]] = {}
                non_exist_targets: set[str] = set()
                for i in q.queries:
                    if i.target in buff:
                        continue
                    else:
                        t_collection: Optional[dict[str, Any]] = self.collection_to_dict(i.target)
                        if t_collection is not None:
                            buff[i.target] = t_collection
                            # コレクションをトランザクションモードに変更する。
                            self.collection(i.target).change_transaction_mode(True)
                        else:
                            non_exist_targets.add(i.target)
                try:
                    for i in q.queries:
                        results.append(self.execute_query(i, collection_permissions=collection_permissions))
                except Exception:
                    _logger.error("execute_transaction_query: Transaction failed", exc_info=True)
                    return self._rollback_collections(buff=buff, non_exist_targets=non_exist_targets)

                # rollback if any query failed
                if any(not r.is_success for r in results):
                    return self._rollback_collections(buff=buff, non_exist_targets=non_exist_targets)

                # commit: notify listeners
                for key in buff.keys():
                    col = self.collection(key)
                    need_callback = getattr(col, "run_notify_listeners_in_transaction", False)
                    col.change_transaction_mode(False)
                    if need_callback:
                        col.notify_listeners()
                return TransactionQueryResult(is_success=True, results=results)
            except Exception:
                _logger.error("execute_transaction_query: Transaction failed", exc_info=True)
                return TransactionQueryResult(is_success=False, results=[], error_message="Unexpected Error")

    def _rollback_collections(self,
                              buff: dict[str, dict[str, Any]],
                              non_exist_targets: set[str],
                              ) -> TransactionQueryResult:
        # DBの変更を元に戻す。
        for key in buff.keys():
            self.collection_from_dict_keep_listener(key, buff[key])
            # 念のため確実に false にする。
            self.collection(key).change_transaction_mode(False)
        # 操作前に存在しなかったコレクションは削除する。
        for key in non_exist_targets:
            self.remove_collection(key)
        _logger.error("execute_transaction_query: Transaction failed", exc_info=True)
        return TransactionQueryResult(
            is_success=False,
            results=[],
            error_message="Transaction failed",
        )



# coding: utf-8
from typing import List, Dict, Any

from file_state_manager.cloneable_file import CloneableFile
from file_state_manager.util_object_hash import UtilObjectHash

from delta_trace_db.query.enum_query_type import EnumQueryType


class Permission(CloneableFile):
    className: str = "Permission"
    version: str = "1"

    def __init__(self, allows: List[EnumQueryType]):
        # allows が None の場合は不正
        super().__init__()
        if allows is None:
            raise ValueError("allows cannot be None")
        self.allows: List[EnumQueryType] = allows

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "Permission":
        allows: List[EnumQueryType] = []
        m_allows: List[str] = src.get("allows", [])
        for i in m_allows:
            enum_val = EnumQueryType[i]
            if enum_val is None:
                raise ValueError("Invalid EnumQueryType")
            allows.append(enum_val)
        return cls(allows)

    def clone(self) -> "Permission":
        return Permission.from_dict(self.to_dict())

    def to_dict(self) -> Dict[str, Any]:
        m_allows: List[str] = [i.name for i in self.allows]
        return {
            "className": self.className,
            "version": self.version,
            "allows": m_allows
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Permission):
            return False
        return self.allows == other.allows

    def __hash__(self) -> int:
        return hash(UtilObjectHash.calc_list(self.allows))



# coding: utf-8
from abc import ABC, abstractmethod
from typing import Any, Dict
from delta_trace_db.query.nodes.enum_node_type import EnumNodeType


class QueryNode(ABC):
    """Base class for query nodes.

    (en) Returns True if the object matches the calculation.
    (ja) 計算と一致するオブジェクトだった場合はTrueを返します。
    """

    @abstractmethod
    def evaluate(self, data: Dict[str, Any]) -> bool:
        """Evaluate the node against a data dictionary."""
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """Convert the object to a dictionary."""
        pass

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "QueryNode":
        """Restore a QueryNode object from a dictionary."""
        # 遅延インポート
        from delta_trace_db.query.nodes.logical_node import AndNode, OrNode, NotNode
        from delta_trace_db.query.nodes.comparison_node import FieldEquals, FieldNotEquals, FieldGreaterThan, \
            FieldLessThan, \
            FieldGreaterThanOrEqual, FieldLessThanOrEqual, FieldMatchesRegex, FieldContains, FieldStartsWith, \
            FieldEndsWith, \
            FieldIn, FieldNotIn
        node_type = EnumNodeType[src["type"]]
        match node_type:
            case EnumNodeType.and_:
                return AndNode.from_dict(src)
            case EnumNodeType.or_:
                return OrNode.from_dict(src)
            case EnumNodeType.not_:
                return NotNode.from_dict(src)
            case EnumNodeType.equals_:
                return FieldEquals.from_dict(src)
            case EnumNodeType.notEquals_:
                return FieldNotEquals.from_dict(src)
            case EnumNodeType.greaterThan_:
                return FieldGreaterThan.from_dict(src)
            case EnumNodeType.lessThan_:
                return FieldLessThan.from_dict(src)
            case EnumNodeType.greaterThanOrEqual_:
                return FieldGreaterThanOrEqual.from_dict(src)
            case EnumNodeType.lessThanOrEqual_:
                return FieldLessThanOrEqual.from_dict(src)
            case EnumNodeType.regex_:
                return FieldMatchesRegex.from_dict(src)
            case EnumNodeType.contains_:
                return FieldContains.from_dict(src)
            case EnumNodeType.in_:
                return FieldIn.from_dict(src)
            case EnumNodeType.notIn_:
                return FieldNotIn.from_dict(src)
            case EnumNodeType.startsWith_:
                return FieldStartsWith.from_dict(src)
            case EnumNodeType.endsWith_:
                return FieldEndsWith.from_dict(src)
